import { Context, Middleware as IMiddleware } from 'koa'
import * as request from 'superagent'
import { diff } from 'deep-object-diff'

export interface ISessionMiddlewareOptions {
  url: string
}

const THIRTY_DAYS = 1000 * 60 * 60 * 24 * 30

const defaultOptions: ISessionMiddlewareOptions = {
  url: 'http://localhost:3100',
}

export function Session (optionsIn?: ISessionMiddlewareOptions): IMiddleware {
  const options = Object.assign({}, defaultOptions, optionsIn)

  async function getSession (sid: string) {
    let response

    try {
      response = await request.get(`${options.url}/api/session/${sid}`)
    } catch (err) {
      response = { body: null, status: err.status || 500 }
    }

    if (response.status === 404) {
      return createSession()
    }

    return response.body
  }

  async function createSession () {
    let response

    try {
      response = await request.post(`${options.url}/api/session`)
    } catch (err) {
      response = { body: null }
    }

    return response.body
  }

  async function updateSession (sid: string, obj: any) {
    let response

    try {
      response = await request.patch(`${options.url}/api/session/${sid}`).send(obj)
    } catch (err) {
      response = { body: null }
    }

    return response.body
  }

  return async function Middleware (ctx: Context, next: () => Promise<any>) {
    const sid = ctx.cookies.get('sid')
    let session: any

    if (sid) {
      session = await getSession(sid)
    } else {
      session = await createSession()
    }

    ctx.session = {
      id: session.id,
      maxAge: session.maxAge,
      user: session.user,
      ...session.session,
    }

    ctx.cookies.set('sid', session.id, {
      maxAge: THIRTY_DAYS,
      secure: process.env.NODE_ENV === 'production',
    })

    await next()

    const { id, maxAge, user, ...rest } = ctx.session
    const ctxSession = { id, maxAge, user, session: rest }
    const difference = diff(session, ctxSession)

    if (Object.keys(difference).length > 0) {
      await updateSession(id, diff(session, ctxSession))
    }
  }
}
